#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <signal.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <X11/Xlib.h>
#include <GL/freeglut.h>
#include "wcam.h"

#define SOCK_NAME	"/tmp/.htrack.sock"

#if defined(__BYTE_ORDER) && __BYTE_ORDER == __BIG_ENDIAN
#define RSHIFT		0
#define GSHIFT		8
#define BSHIFT		16
#define PIXFMT	GL_RGBA
#else	/* default to little endian */
#define RSHIFT		16
#define GSHIFT		8
#define BSHIFT		0
#define PIXFMT	GL_BGRA
#endif

#define UNP_RED(x)		(((x) >> RSHIFT) & 0xff)
#define UNP_GREEN(x)	(((x) >> GSHIFT) & 0xff)
#define UNP_BLUE(x)		(((x) >> BSHIFT) & 0xff)

#define PACK_RGB(r, g, b)	\
	((((r) & 0xff) << RSHIFT) | \
	 (((g) & 0xff) << GSHIFT) | \
	 (((b) & 0xff) << BSHIFT))

struct rect {
	int left, right;
	int top, bottom;
};

void init_camera_controls(void);
void redraw(void);
void keyb(unsigned char key, int x, int y);
void mouse_button(int bn, int state, int x, int y);
void mouse_wheel(int unk, int dir, int x, int y);

void draw_rect(struct rect *r);
int is_blank_row(int row, int sx, int ex);
int is_blank_column(int col, int sy, int ey);
int process_img(unsigned int *img, int xsz, int ysz, unsigned int *dest_img);
float color_dist(int *c1, int *c2);
void print_string(void *font, const char *str);
void capture_frame(const char *prefix);
void post_redisplay(void);

int xsz, ysz;
int need_redisplay = 1;

int cam_fd;
unsigned int *pixels, *original, *scrap;
int pixels_valid;

int tracking;
int capture;
/*int exposure = 340;*/
int exposure = 800;

int no_redraw;

int min_exp, max_exp;

int track_color[3] = {255, 255, 255};
float threshold = 0.03;
struct rect brect, left_rect, right_rect;

int lis = -1;	/* listening PF_UNIX socket */
int conn[256];
int num_conn;

/* hack, see freeglut_internal.h and freeglut_init.c */
extern Display *fgDisplay;
int x11_sock;

int main(int argc, char **argv)
{
	struct sockaddr_un addr;
	mode_t prev_umask;

	glutInitWindowSize(640, 480);
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);
	glutCreateWindow("camera test");
	xsz = glutGet(GLUT_WINDOW_WIDTH);
	ysz = glutGet(GLUT_WINDOW_HEIGHT);

	glutDisplayFunc(redraw);
	glutKeyboardFunc(keyb);
	glutMouseFunc(mouse_button);
	glutMouseWheelFunc(mouse_wheel);

	x11_sock = ConnectionNumber(fgDisplay);

	/* initialize webcamera */
	if((cam_fd = wcam_open(0, xsz, ysz, 25)) == -1) {
		return 1;
	}
	init_camera_controls();

	/* create UNIX domain socket that listens for  */
	if((lis = socket(PF_UNIX, SOCK_STREAM, 0)) == -1) {
		perror("failed to create socket");
		return 1;
	}
	unlink(SOCK_NAME);

	memset(&addr, 0, sizeof addr);
	addr.sun_family = AF_UNIX;
	strcpy(addr.sun_path, SOCK_NAME);

	prev_umask = umask(0);
	if(bind(lis, (struct sockaddr*)&addr, SUN_LEN(&addr)) == -1) {
		perror("failed to bind socket: " SOCK_NAME);
		return 1;
	}
	umask(prev_umask);
	listen(lis, 32);

	/* allocate framebuffers */
	if(!(pixels = malloc(xsz * ysz * sizeof *pixels))) {
		perror("failed to allocate frame buffer");
		return 1;
	}
	if(!(original = malloc(xsz * ysz * sizeof *original))) {
		perror("failed to allocate output image");
		return 1;
	}
	if(!(scrap = malloc(xsz * ysz * sizeof *scrap))) {
		perror("failed to allocate memory");
		return 1;
	}

	/* initialize OpenGL */
	glMatrixMode(GL_PROJECTION);
	glOrtho(0, xsz, 0, ysz, -1, 1);
	glMatrixMode(GL_MODELVIEW);

	glClearColor(0, 0, 0, 0);

	signal(SIGPIPE, SIG_IGN);	/* ignore SIGPIPE, in order to get EPIPE on write(2) */

	wcam_start(cam_fd);

	for(;;) {
		int maxfd;
		fd_set rdset;

		FD_ZERO(&rdset);
		FD_SET(x11_sock, &rdset);
		FD_SET(lis, &rdset);
		FD_SET(cam_fd, &rdset);

		maxfd = x11_sock | lis | cam_fd;

		if(select(maxfd + 1, &rdset, 0, 0, 0) == -1) {
			if(errno != EINTR) {
				perror("select failed");
				break;
			}
			continue;
		}

		/* got frame? */
		if(FD_ISSET(cam_fd, &rdset)) {
			wcam_read_frame_rgb(cam_fd, original);

			if(tracking) {
				int i, x1, y1, x2, y2;
				float msg[4];

				process_img(original, xsz, ysz, pixels);

				x1 = (left_rect.left + left_rect.right) / 2;
				y1 = (left_rect.top + left_rect.bottom) / 2;
				x2 = (right_rect.left + right_rect.right) / 2;
				y2 = (right_rect.top + right_rect.bottom) / 2;

				msg[0] = (float)x1 / xsz;
				msg[1] = (float)y1 / ysz;
				msg[2] = (float)x2 / xsz;
				msg[3] = (float)y2 / ysz;

				for(i=0; i<num_conn; i++) {
					if(write(conn[i], msg, sizeof msg) == -1) {
						if(errno == EPIPE) {
							printf("client %d disconnected\n", i);
							memmove(conn + i, conn + i + 1, (num_conn - i - 1) * sizeof *conn);
							num_conn--;
							i--;	/* so that the new fd in slot i won't be skipped over */
						} else if(errno != EINTR) {
							perror("failed to send event");
						}
					}
				}
			}
			post_redisplay();
		}

		/* got client? */
		if(FD_ISSET(lis, &rdset)) {
			int s;
			if((s = accept(lis, 0, 0)) == -1) {
				perror("failed to establish connection");
			}
			printf("new client: %d\n", num_conn);
			conn[num_conn++] = s;
		}

		/* got X event? */
		if(FD_ISSET(x11_sock, &rdset) || need_redisplay) {
			glutMainLoopEvent();
		}
	}

	unlink(SOCK_NAME);
	wcam_close(cam_fd);
	return 0;
}

void init_camera_controls(void)
{
	int low, high, step;

	if(wcam_check_ctl(cam_fd, WCAM_CTL_AUTO_EXPOSURE, &low, &high, &step) == 0) {
		printf("auto exposure: [%d, %d] step %d\n", low, high, step);
		wcam_set_ctl(cam_fd, WCAM_CTL_AUTO_EXPOSURE, WCAM_AUTOEXP_OFF);
	} else {
		printf("warning: can't control auto-exposure. if it's available it might cause problems\n");
	}

	if(wcam_check_ctl(cam_fd, WCAM_CTL_EXPOSURE_ABS, &min_exp, &max_exp, &step) == 0) {
		printf("exposure abs: [%d, %d] step %d\n", min_exp, max_exp, step);
		wcam_set_ctl(cam_fd, WCAM_CTL_EXPOSURE_ABS, exposure);
	} else {
		if(wcam_check_ctl(cam_fd, WCAM_CTL_EXPOSURE, &min_exp, &max_exp, &step) == 0) {
			printf("exposure: [%d, %d] step %d\n", min_exp, max_exp, step);
			exposure = wcam_get_ctl(cam_fd, WCAM_CTL_EXPOSURE);
		} else {
			printf("warning: can't control exposure\n");
		}
	}

	if(wcam_check_ctl(cam_fd, WCAM_CTL_AUTO_WHITE_BALANCE_TEMP, &low, &high, &step) == 0) {
		printf("auto white balance temperature: [%d, %d] step %d\n", low, high, step);
		wcam_set_ctl(cam_fd, WCAM_CTL_AUTO_WHITE_BALANCE_TEMP, low);
	} else {
		if(wcam_check_ctl(cam_fd, WCAM_CTL_AUTO_WHITE_BALANCE, &low, &high, &step) == 0) {
			printf("auto white balance: [%d, %d] step %d\n", low, high, step);
			wcam_set_ctl(cam_fd, WCAM_CTL_AUTO_WHITE_BALANCE, low);
		} else {
			printf("warning: can't control auto-white-balance. if it's available it might cause problems\n");
		}
	}
}

void redraw(void)
{
	if(no_redraw) {
		return;
	}

	glClear(GL_COLOR_BUFFER_BIT);

	glRasterPos2i(0, ysz - 1);
	glPixelZoom(1, -1);
	glDrawPixels(xsz, ysz, PIXFMT, GL_UNSIGNED_BYTE, original);
	glPixelZoom(1, 1);

	if(tracking) {
		char tmp_str[64];

		glColor3f(0.2, 0.3, 1.0);
		draw_rect(&left_rect);
		draw_rect(&right_rect);

		sprintf(tmp_str, "thres: %.3f", threshold);
		glRasterPos2i(5, 10);
		print_string(GLUT_BITMAP_HELVETICA_18, tmp_str);

		sprintf(tmp_str, "track_color: %d %d %d", track_color[0], track_color[1], track_color[2]);
		glRasterPos2i(5, 30);
		print_string(GLUT_BITMAP_HELVETICA_18, tmp_str);
	}

	if(capture) {
		capture_frame("frame");
	}

	glutSwapBuffers();
}

void keyb(unsigned char key, int x, int y)
{
	switch(key) {
	case 27:
		exit(0);

	case ' ':
		capture_frame("shot");
		printf("picture taken\n");
		/*
		capture = !capture;
		printf("capture: %s\n", capture ? "ON" : "OFF");
		*/
		break;

	case 't':
	case 'T':
		tracking = !tracking;
		printf("tracking %s\n", tracking ? "enabled" : "disabled");
		break;

	case 'n':
		no_redraw = !no_redraw;
		break;

	default:
		break;
	}
}

void mouse_button(int bn, int state, int x, int y)
{
	if(bn == GLUT_LEFT_BUTTON && state == GLUT_UP) {
		unsigned int pixel = original[y * xsz + x];

		track_color[0] = UNP_RED(pixel);
		track_color[1] = UNP_GREEN(pixel);
		track_color[2] = UNP_BLUE(pixel);

		post_redisplay();
	}
}

void mouse_wheel(int unk, int dir, int x, int y)
{
	if(tracking) {
		threshold += dir > 0 ? 0.01 : -0.01;

		if(threshold > 1.0) threshold = 1.0;
		if(threshold < 0.0) threshold = 0.0;
	} else {
		exposure += dir > 0 ? 5 : -5;

		if(exposure < min_exp) exposure = min_exp;
		if(exposure > max_exp) exposure = max_exp;

		wcam_set_ctl(cam_fd, WCAM_CTL_EXPOSURE_ABS, exposure);
		printf("exposure: %d    \r", exposure);
		fflush(stdout);
	}
	post_redisplay();
}

/* -------------------------------------------------------------------- */

void draw_rect(struct rect *r)
{
	int cx, cy;

	glLineWidth(2);
	glBegin(GL_LINE_LOOP);
	glColor3f(0, 0, 1);
	glVertex2i(r->left, ysz - r->top);
	glVertex2i(r->right, ysz - r->top);
	glVertex2i(r->right, ysz - r->bottom);
	glVertex2i(r->left, ysz - r->bottom);
	glEnd();

	cx = (r->left + r->right) / 2;
	cy = (r->top + r->bottom) / 2;

	glBegin(GL_LINES);
	glColor3f(1, 1, 0);
	
	glVertex2i(cx, ysz - r->top);
	glVertex2i(cx, ysz - r->top + 10);

	glVertex2i(cx, ysz - r->bottom);
	glVertex2i(cx, ysz - r->bottom - 10);

	glVertex2i(r->left, ysz - cy);
	glVertex2i(r->left - 10, ysz - cy);

	glVertex2i(r->right, ysz - cy);
	glVertex2i(r->right + 10, ysz - cy);
	glEnd();
}

int is_blank_row(int row, int sx, int ex)
{
	int i;
	unsigned int *ptr = scrap + row * xsz + sx;

	for(i=sx; i<ex; i++) {
		if(*ptr++ > 0) {
			return 0;
		}
	}
	return 1;
}

int is_blank_column(int col, int sy, int ey)
{
	int i;
	unsigned int *ptr = scrap + sy * xsz + col;

	for(i=sy; i<ey; i++) {
		if(*ptr > 0) {
			return 0;
		}
		ptr += xsz;
	}
	return 1;
}

int process_img(unsigned int *img, int xsz, int ysz, unsigned int *dest_img)
{
	int i, numpix;

	brect.left = brect.top = 0;
	brect.right = xsz;
	brect.bottom = ysz;

	numpix = xsz * ysz;

	/* keep only closest-colored pixels */
	for(i=0; i<numpix; i++) {
		int pixel[3];

		pixel[0] = UNP_RED(img[i]);
		pixel[1] = UNP_GREEN(img[i]);
		pixel[2] = UNP_BLUE(img[i]);

		scrap[i] = color_dist(pixel, track_color) < threshold ? img[i] : 0;
	}

	/* find first-level bounding box (left/right only) */
	for(i=0; i<xsz; i++) {
		if(!is_blank_column(brect.left++, 0, ysz)) {
			break;
		}
	}
	for(i=0; i<xsz; i++) {
		if(brect.right <= brect.left || !is_blank_column(--brect.right, 0, ysz)) {
			break;
		}
	}

	/* find second-level bounding boxen ... */
	left_rect = right_rect = brect;
	left_rect.right = right_rect.left = (brect.right + brect.left) / 2;

	while(left_rect.right > left_rect.left && is_blank_column(left_rect.right, 0, ysz)) {
		left_rect.right--;
	}

	while(right_rect.left < right_rect.right && is_blank_column(right_rect.left, 0, ysz)) {
		right_rect.left++;
	}

	/* also shrink along the Y axis */
	left_rect.bottom--;
	left_rect.top++;
	while(left_rect.top < left_rect.bottom
			&& is_blank_row(left_rect.top, left_rect.left, left_rect.right)) {
		left_rect.top++;
	}
	while(left_rect.bottom > left_rect.top
			&& is_blank_row(left_rect.bottom, left_rect.left, left_rect.right)) {
		left_rect.bottom--;
	}

	right_rect.bottom--;
	while(right_rect.top < right_rect.bottom
			&& is_blank_row(right_rect.top, right_rect.left, right_rect.right)) {
		right_rect.top++;
	}
	while(right_rect.bottom > right_rect.top
			&& is_blank_row(right_rect.bottom, right_rect.left, right_rect.right)) {
		right_rect.bottom--;
	}


	/* blah blah */
	/*for(i=0; i<numpix; i++) {
		dest_img[i] = scrap[i] ? 0x00ff00 : img[i];
	}*/
	return 0;
}

#define SQ(x)	((x) * (x))
float color_dist(int *c1, int *c2)
{
	float dr, dg, db, foo;
	static float max_diff = -1.0;

	if(max_diff < 0.0) {
		max_diff = sqrt(3.0 + 4.0 + 2.0);
	}
	
	dr = (c1[0] - c2[0]) / 255.0;
	dg = (c1[1] - c2[1]) / 255.0;
	db = (c1[2] - c2[2]) / 255.0;
	/*foo = (float)sqrt(2.0 * SQ(dr) + 4.0 * SQ(dg) + 3.0 * SQ(db)) / max_diff;*/
	foo = (2.0 * SQ(dr) + 4.0 * SQ(dg) + 3.0 * SQ(db)) / max_diff;
	if(foo > 0) {
		/*printf("%f\n", foo);*/
	}
	return foo;
}

void print_string(void *font, const char *str)
{
	int rpos[4];

	glColor3i(255, 255, 255);

	glGetIntegerv(GL_CURRENT_RASTER_POSITION, rpos);
	while(*str) {
		glutBitmapCharacter(font, *str);
		rpos[0] += glutBitmapWidth(font, *str);
		glRasterPos2i(rpos[0], rpos[1]);
		str++;
	}
}

void capture_frame(const char *prefix)
{
	static int scr_num[256];
	static char fname[50];
	int i, size = xsz * ysz;
	FILE *fp;

	sprintf(fname, "%s%04d.ppm", prefix, scr_num[(int)*prefix]++);

	glReadPixels(0, 0, xsz, ysz, PIXFMT, GL_UNSIGNED_BYTE, pixels);

	if((fp = fopen(fname, "wb"))) {
		fprintf(fp, "P6\n%d %d\n255\n", xsz, ysz);

		for(i=0; i<size; i++) {
			fputc((pixels[i] >> RSHIFT) & 0xff, fp);
			fputc((pixels[i] >> GSHIFT) & 0xff, fp);
			fputc((pixels[i] >> BSHIFT) & 0xff, fp);
		}

		fclose(fp);
	}
}

void post_redisplay(void)
{
	need_redisplay = 1;
	glutPostRedisplay();
}
